

from tendril.libraries import interests
from tendril.common.interests.exceptions import InterestTypeUnsupported

from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__)


@with_db
def request_device_logs(id: int, background_tasks=None, auth_user=None, session=None):
    library = interests.find_library(id)
    interest = library.item(id=id, session=session)
    if not hasattr(interest, 'appname'):
        raise InterestTypeUnsupported("'appname' attribute", id=id, name=interest.name)
    interest.trigger_device_publish_logs(background_tasks=background_tasks)
    return 'OK'


@with_db
def available_device_logs(id: int, background_tasks=None, auth_user=None, session=None):
    library = interests.find_library(id)
    interest = library.item(id=id, session=session)
    if not hasattr(interest, 'appname'):
        raise InterestTypeUnsupported("'appname' attribute", id=id, name=interest.name)
    # TODO find logs in filestore and return links
    raise NotImplementedError



# Copyright (C) 2023 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Interests Configuration Options
===============================
"""


from tendril.utils.config import ConfigOption
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)

depends = ['tendril.config.core']

config_elements_iotedge = [
    ConfigOption(
        'IOTEDGE_API_ENABLED',
        "False",
        "Whether the IOT Edge API is enabled on this instance / component. Generally, "
        "multi-component Tendril deployments will have the IOT Edge API enabled only on "
        "a single component. For such deployments, this parameter is generally best set through "
        "environment variables.",
        parser=bool
    ),
    ConfigOption(
        'IOTEDGE_ANNOUNCE_ENDPOINT_OPEN',
        "True",
        "Whether the /announce endpoint is open (unauthenticated). This is required for the "
        "integrated device registration flow.",
        parser=bool
    ),
    ConfigOption(
        'IOTEDGE_DEVICE_LOGS_UPLOAD_FILESTORE_BUCKET',
        '"incoming"',
        "Filestore bucket to upload device logs to."
    )
]


def load(manager):
    logger.debug("Loading {0}".format(__name__))
    manager.load_elements(config_elements_iotedge,
                          doc="Tendril IOT Edge Configuration")





import os
import arrow
from httpx import HTTPStatusError
from tendril.filestore import buckets
from tendril.config import IOTEDGE_DEVICE_LOGS_UPLOAD_FILESTORE_BUCKET

from tendril.caching import tokens
from tendril.caching.tokens import TokenStatus
from tendril.db.models.deviceconfig import DeviceConfigurationModel
from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class DeviceProfile(object):
    appname = None
    interest_type = 'device'
    config_model = DeviceConfigurationModel
    logs_upload_bucket_name = IOTEDGE_DEVICE_LOGS_UPLOAD_FILESTORE_BUCKET

    def __init__(self, model_instance):
        self._model_instance = model_instance

    @property
    def upload_bucket(self):
        if not hasattr(self, '_upload_bucket'):
            self._logs_upload_bucket = None
        if not self._logs_upload_bucket:
            self._logs_upload_bucket = buckets.get_bucket(self.logs_upload_bucket_name)
        return self._logs_upload_bucket

    @property
    def model_instance(self):
        return self._model_instance

    def interest(self):
        from tendril import interests
        itype = interests.type_codes[self.interest_type]
        if isinstance(self.model_instance, itype):
            return self.model_instance
        rv = itype(self.model_instance)
        return rv

    def report_seen(self, background_tasks=None):
        self.interest().monitor_report('last_seen', arrow.utcnow(), background_tasks=background_tasks)
        self.interest().monitor_report('online', True, background_tasks=background_tasks)

    def report_status(self, status, background_tasks=None):
        self.interest().monitors_report(status,
                                        background_tasks=background_tasks)

    def _report_filestore_error(self, token_id, e, action_comment):
        logger.warn(f"Exception while {action_comment} : HTTP {e.response.status_code} {e.response.text}")
        if token_id:
            tokens.update(
                'dlu', token_id, state=TokenStatus.FAILED,
                error={"summary": f"Exception while {action_comment}",
                       "filestore": {
                           "code": e.response.status_code,
                           "content": e.response.json()}
                       }
            )

    @with_db
    async def receive_logs(self, file, rename_to=None, token_id=None, session=None):
        # TODO Consider integrating this into a standardized upload handler
        #  in the interest itself
        storage_folder = f'{self.model_instance.id}'
        if token_id:
            tokens.update('dlu', token_id,
                          state=TokenStatus.INPROGRESS, max=2,
                          current="Uploading File to Filestore")

        filename = rename_to or file.filename

        # 2. Upload File to Bucket
        try:
            upload_response = await self.upload_bucket.upload(
                file=(os.path.join(storage_folder, filename), file.file),
                interest=self.model_instance.id, label="device_logs"
            )
        except HTTPStatusError as e:
            self._report_filestore_error(token_id, e, "uploading device logs file to bucket")
            return

        if token_id:
            tokens.update('dlu', token_id, current="Finishing",  done=2,
                          metadata={'storedfile_id': upload_response['storedfileid']}, )

        # 7. Close Upload Ticket
        tokens.close('dlu', token_id)

    @with_db
    def config(self, session=None):
        if hasattr(self.model_instance, 'model_instance'):
            logger.warn("model_instance is set to something that isn't a model instance.")
            model_instance = self.model_instance.model_instance
        else:
            model_instance = self.model_instance
        session.add(model_instance)
        cfg = model_instance.config
        if not cfg:
            cfg = self.config_model()
            model_instance.config = cfg
            session.add(cfg)
            session.flush()
        return cfg

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.model_instance.appname} {self.model_instance.name}>"


def load(manager):
    pass





from tendril.caching import transit
from tendril.common.states import LifecycleStatus

from .registration import get_registration
from .registration import register

from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


@with_db
def announce_device(device_id, appname, have_credentials,
                    background_tasks, device_sysinfo=None, session=None):
    # TODO Handle other device lifecycle states
    rv = {}
    device = get_registration(device_id, appname, session=session)
    if device:
        logger.debug(f"Got announce from a registered device {device_id} "
                     f"in state {device.model_instance.status}.")
        if device.model_instance.status == LifecycleStatus.ACTIVE:
            if not have_credentials:
                # If there is a password waiting for one-time transmission,
                # transmit it.
                password = transit.read(namespace="ott:dp", key=device_id)
                if not password:
                    logger.warning(f"Active registered device {device_id} does not have a token")
                    # If there is no password waiting for one-time transmission,
                    # some user should check the request is not spurious, and then
                    # manually reset password and load it up for one-time transmission.
                    # Raise manual password reset request here. Changing device status to
                    # NEW should get the job done as presently implemented.
                    # TODO This needs to be implemented on priority. Switching SD
                    #  cards on devices will bring us here.
                else:
                    logger.info(f"Found password for {device_id} on transit cache.")
                    transit.delete(namespace="ott:dp", key=device_id)
                    rv['password'] = password
    else:
        logger.info(f"Registering new device '{device_id}' "
                    f"of type '{appname}' from announce.")
        if have_credentials:
            logger.warning(f"Unregistered device {device_id} seems to already have a token")
            # Raise some kind of error here. The device might need to checked for compromise.
        device = register(device_id, appname, info={}, session=session)
        # When the device is activated, the interest activation will create a password
        # and store it on transit for one time transmission within this endpoint itself.
    rv.update({
        'interest_id': device.model_instance.id,
        'id': device.model_instance.name,
        'status': device.model_instance.status
    })
    if device_sysinfo:
        background_tasks.add_task(device.interest().replace_info,
                                  info=device_sysinfo,
                                  section='sysinfo',
                                  context='announce',
                                  ignore_changes=["root['sysinfo']['status']",
                                                  "root['host']['boot_id']"])
    return rv





import os
import datetime
from tendril.caching import tokens

from .registration import registered_device

from tendril.common.exceptions import FileTypeNotSupportedException
from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__)


@with_db
@registered_device()
async def publish_logs(logs_archive, device=None, appname=None,
                       background_tasks=None, session=None):
    logger.info(f"Got logs from {device}")
    file_name, file_ext = os.path.splitext(logs_archive.filename)
    if file_ext not in ['.zip']:
        raise FileTypeNotSupportedException

    device_name = device.model_instance.name
    ts = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    storage_filename = f"{device_name}.{ts}.logs{file_ext}"

    upload_token = tokens.open(
        namespace='dlu',
        metadata={'device': device_name,
                  'filename': storage_filename},
        current="Request Created",
        progress_max=1, ttl=600,
    )

    background_tasks.add_task(
        device.recieve_logs, file=logs_archive,
        rename_to=storage_filename, token_id=upload_token.id,
        session=session
    )

    return upload_token

